//
// Copyright (c) 2018-2020 CNRS INRIA
//

#ifndef __pinocchio_multibody_liegroup_liegroup_algo_hpp__
#define __pinocchio_multibody_liegroup_liegroup_algo_hpp__

#include "pinocchio/multibody/joint/joint-base.hpp"
#include "pinocchio/multibody/liegroup/liegroup.hpp"

#include "pinocchio/multibody/liegroup/liegroup-algo.hxx"

#endif // ifndef __pinocchio_multibody_liegroup_liegroup_algo_hpp__
